%% Camera tracking using direct image alignment
clear all; close all; clc;

data_dir = '/home/zeni0001/Desktop/rgbd_dataset_freiburg2_desk';
match_file_path = 'matches_fr2_desk.txt';
K = [520.9 0 325.1;	0 521.0 249.7; 0 0 1]; %fr2

% Read files names
fid = fopen(match_file_path);
nb = 0;
filenames = cell(3000, 2);

while ~feof(fid)
    nb = nb + 1;
    line = fgetl(fid); 
    filenames(nb, 1:2) = textscan(line,'%s %s'); 
end

fclose(fid);
filenames = filenames(1:nb, :);

height = 480;
width = 640;

% Pre-allocate memory
% Relative poses to reference keyframe
Rs_kc = zeros(3, 3, nb);
Ts_kc = zeros(3, nb);

% Reference frame id
ref_ids = zeros(nb, 1);

% Only store images for KFs
kf_rgb_imgs = zeros(height, width, 1);
kf_depth_imgs = zeros(height, width, 1);

% Global poses of KFs
kf_Rs_wk = zeros(3, 3, 1);
kf_Ts_wk = zeros(3, 1);
kf_ids = zeros(nb, 1);

% Set first frame
kf_Rs_wk(:, :, 1) = eye(3);
kf_Ts_wk(:, 1) = zeros(3, 1);
kf_rgb_imgs(:, :, 1) = double(imreadbw([data_dir '/' filenames{1, 1}{1}]));
kf_depth_imgs(:, :, 1) = double(imread([data_dir '/' filenames{1, 2}{1}]))/5000;
ref_ids(1) = 1;
Rs_kc(:, :, 1) = eye(3);
Ts_kc(:, 1) = zeros(3, 1);
kf_ids(1) = 1;

huber_delta = 0.2;
curr_kf_id = 1;
start_level = 5;
end_level = 2;
threshold = 0.003;
weights_new_kf = ones(6,1);
kf_num = 1;

for i = 2:nb
    fprintf('Track frame %d on %d (kf %d)\n', i, kf_ids(kf_num), kf_num);
    
    % Read new image
    rgb_img = double(imreadbw([data_dir '/' filenames{i, 1}{1}]));
    depth_img = double(imread([data_dir '/' filenames{i, 2}{1}]))/5000;
    
    % initialization
    xi = [0 0 0 0 0 0]';
    
    % pyramid levels
    for lvl = start_level:-1:end_level
        % get downscaled image, depth image, and K-matrix of down-scaled image.
        [IRef, DRef, Klvl] = downscale(kf_rgb_imgs(:, :, kf_num), kf_depth_imgs(:, :, kf_num), K, lvl);
        [I, D] = downscale(rgb_img, depth_img, K, lvl);
        
        % just do at most 20 steps.
        errLast = 1e10;
        for iter=1:20
            % calculate Jacobian of residual function (Matrix of dim (width*height) x 6)
            [Jac, residual, weights] = deriveResidualsAnalytic(IRef,...
                DRef, I, xi, Klvl, huber_delta);
            
            % set rows with NaN to 0 (e.g. because out-of-bounds or invalid depth).
            notValid = isnan(sum(Jac,2)+residual);
            residual(notValid,:) = 0;
            Jac(notValid,:) = 0;
            weights(notValid,:) = 0;

            % do Gauss-Newton step
            upd = - (Jac' * (repmat( weights, 1, 6 ) .* Jac))^-1 * Jac'...
                * (weights .* residual);
            
            % multiply increment from left onto the current estimate
            lastXi = xi;
            xi = se3Log(se3Exp(upd) * se3Exp(xi));
            
            err = mean(residual .* residual);
            
            if(err / errLast > 0.995)
                break;
            end   
            
            errLast = err;
        end
    end
    
    ref_ids(i) = kf_num;
    
    % Store results
    T_kc = inv(se3Exp(xi));
    Rs_kc(:, :, i) = T_kc(1:3, 1:3);
    Ts_kc(:, i) = T_kc(1:3, 4);
       
    % Check if new KF is needed
    dist = (weights_new_kf .* xi)' * xi;
    if (dist > threshold)
        kf_num = kf_num + 1;
        
        % Store images
        kf_rgb_imgs(:, :, kf_num) = rgb_img;
        kf_depth_imgs(:, :, kf_num) = depth_img;
        
        % Calculate and store global pose
        T_wk = [kf_Rs_wk(:, :, ref_ids(i)) kf_Ts_wk(:, ref_ids(i));...
            0 0 0 1] * T_kc;
        kf_Rs_wk(:, :, kf_num) = T_wk(1:3, 1:3);
        kf_Ts_wk(:, kf_num) = T_wk(1:3, 4);

        kf_ids(kf_num) = i;
    end
end

%kf_rgb_imgs = kf_rgb_imgs(:, :, 1:kf_num);
%kf_depth_imgs = kf_depth_imgs(:, :, 1:kf_num);
kf_Rs_wk = kf_Rs_wk(:, :, 1:kf_num);
kf_Ts_wk = kf_Ts_wk(:, 1:kf_num);
kf_ids = kf_ids(1:kf_num);

save('fr2_desk_5_2_Rs_kc.mat', 'Rs_kc');
save('fr2_desk_5_2_Ts_kc.mat', 'Ts_kc');
save('fr2_desk_5_2_ref_ids.mat', 'ref_ids');
save('fr2_desk_5_2_kf_Rs_wk.mat', 'kf_Rs_wk');
save('fr2_desk_5_2_kf_Ts_wk.mat', 'kf_Ts_wk');
save('fr2_desk_5_2_filenames.mat', 'filenames');
save('fr2_desk_5_2_kf_rgb_imgs.mat', 'kf_rgb_imgs');
save('fr2_desk_5_2_kf_depth_imgs.mat', 'kf_depth_imgs');
save('fr2_desk_5_2_kf_ids.mat', 'kf_ids');
save('fr2_desk_5_2_K.mat', 'K');

% Output file for evaluation
fname = sprintf('fr2_desk_level_%d_%d.txt', start_level, end_level);
fid = fopen(fname, 'w');
for i = 1:1:nb
    ref_id = ref_ids(i);
    T_wk = [kf_Rs_wk(:, :, ref_id) kf_Ts_wk(:, ref_id); 0 0 0 1];
    T_kc = [Rs_kc(:, :, i) Ts_kc(:, i); 0 0 0 1];
    T_wc = T_wk * T_kc;
    quat = rotm2quat(T_wc(1:3, 1:3));
    timestamp = textscan(filenames{i, 2}{1}, 'depth/%f.png');
    fprintf(fid,'%.6f %f %f %f %f %f %f %f\n', timestamp{1}, T_wc(1, 4),...
        T_wc(2, 4),T_wc(3, 4), quat(1), quat(2), quat(3), quat(4));
end
fclose(fid);