%% Pose-graph optimization
clear all; close all; clc;

huber_delta = 0.2;
curr_kf_id = 1;
start_level = 5;
end_level = 2;
threshold = 0.003;

% Load saved data
files = dir('*.mat'); for i = 1:numel(files) load(files(i).name); end

kf_num = size(kf_rgb_imgs,3);

% Manually detect loop
ref_id = 392;
curr_id = 1;  

% Estimate relative pose using direct image alignment    
% initialization
xi_loop = [0 0 0 0 0 0]';
% pyramid levels
for lvl = start_level:-1:end_level
    % get downscaled image, depth image, and K-matrix of down-scaled image.
    [IRef, DRef, Klvl] = downscale(kf_rgb_imgs(:, :, ref_id),...
        kf_depth_imgs(:, :, ref_id), K, lvl);
    [I, D] = downscale(kf_rgb_imgs(:, :, curr_id), ...
        kf_depth_imgs(:, :, curr_id), K, lvl);
    
    % just do at most 20 steps.
    errLast = 1e10;
    for iter=1:20
        % calculate Jacobian of residual function (Matrix of dim (width*height) x 6)
        [Jac, residual, weights] = deriveResidualsAnalytic(IRef,...
            DRef, I, xi_loop, Klvl, huber_delta);
        
        % set rows with NaN to 0 (e.g. because out-of-bounds or invalid depth).
        notValid = isnan(sum(Jac,2)+residual);
        residual(notValid,:) = 0;
        Jac(notValid,:) = 0;
        weights(notValid,:) = 0;
        
        % do Gauss-Newton step
        upd = - (Jac' * (repmat( weights, 1, 6 ) .* Jac))^-1 * Jac'...
            * (weights .* residual);
        
        % multiply increment from left onto the current estimate
        xi_loop = se3Log(se3Exp(upd) * se3Exp(xi_loop));
        
        err = mean(residual .* residual);
        
        if(err / errLast > 0.995)
            break;
        end
        
        errLast = err;
    end
end

% Show residual
figure;
imagesc(reshape(residual,size(kf_rgb_imgs(:, :, curr_id))/end_level));
colormap gray;
set(gca, 'CLim', [-1,1]);

% Plot trajectory before loop closure
figure;
plot3(kf_Ts_wk(1, :), kf_Ts_wk(2, :), kf_Ts_wk(3, :), 'b');
axis equal;
xlabel('x');
ylabel('y');
zlabel('z');
hold on;

% Gauss-Newton
% Initial state
Xi = zeros(kf_num * 6, 1);
for i = 1:kf_num
   xi = se3Log([kf_Rs_wk(:, :, i) kf_Ts_wk(:, i); 0 0 0 1]);
   Xi(6 * i - 5:6 * i) = xi;
end

errLast = 1e10;
for iter = 1:20
    fprintf('Loop closure iteration: %d\n', iter);
    % calculate Jacobian of residual function
    % Calculate residual
    Res = zeros(kf_num * 6, 1);
    for i = 1:kf_num - 1
        T_ji = inv([Rs_kc(:, :, kf_ids(i+1)) Ts_kc(:, kf_ids(i+1)); 0 0 0 1]);
        T_wi = se3Exp(Xi(6 * i - 5:6 * i));
        T_wj = se3Exp(Xi(6 * i + 1:6 * i + 6));
        Res(6 * i - 5:6 * i) = se3Log(inv(T_wi) * T_wj * T_ji);
    end
    % The part for loop contraints
    T_ji = se3Exp(xi_loop);
    T_wi = se3Exp(Xi(6 * ref_id - 5:6 * ref_id));
    T_wj = se3Exp(Xi(6 * curr_id - 5:6 * curr_id));
    Res(end - 5:end) = se3Log(inv(T_wi) * T_wj * T_ji);
    
    % Build up Jacobian
    J = zeros(kf_num * 6);
    for i = 1:kf_num - 1
        T_wi = se3Exp(Xi(6 * i - 5:6 * i));
        Ad = adjoint(inv(T_wi));
        J(6 * i - 5:6 * i, 6 * i - 5:6 * i) = -Ad;
        J(6 * i - 5:6 * i, 6 * i + 1:6 * i + 6) = Ad;
    end
    % Jacobian for the loop constraint
    T_wi = se3Exp(Xi(6 * ref_id - 5:6 * ref_id));
    Ad = adjoint(inv(T_wi));
    J(end - 5:end, 6 * curr_id - 5:6 * curr_id) = Ad;
    J(end - 5:end, 6 * ref_id - 5:6 * ref_id) = -Ad;
    
    % Fix the first KF
    J = J(:, 7:end);
    
    % do Gauss-Newton step
    upd = - inv(J' * J) * J' * Res;
    
    % multiply increment from left onto the current estimate
    for i = 1:kf_num - 1
        delta = upd(6 * i - 5:6 * i);
        xi = Xi(6 * i + 1:6 * i + 6);
        T_wi_new = se3Exp(delta) * se3Exp(xi);
        Xi(6 * i + 1:6 * i + 6) = se3Log(T_wi_new);
    end   
    
    err = Res' * Res
    
    if(err / errLast > 0.995)
        break;
    end
    
    errLast = err;
end

% store optimized poses
% translation
kf_Ts_wk_new = zeros(3, kf_num);
% rotation
kf_Rs_wk_new = zeros(3, 3, kf_num);

for i = 1:kf_num
    T_wi_new = se3Exp(Xi(6 * i -5:6 * i))
    kf_Ts_wk_new(:, i) = T_wi_new(1:3, 4);
    kf_Rs_wk_new(:, :, i) =  T_wi_new(1:3, 1:3);
end

plot3(kf_Ts_wk_new(1, :), kf_Ts_wk_new(2, :), kf_Ts_wk_new(3, :), 'r');

% Output file for evaluation
fname = sprintf('fr2_desk_level_%d_%d_loopclosure.txt', start_level, end_level);
fid = fopen(fname, 'w');
for i = 1:length(ref_ids)
    ref_id = ref_ids(i);
    T_wk = [kf_Rs_wk_new(:, :, ref_id) kf_Ts_wk_new(:, ref_id); 0 0 0 1];
    T_kc = [Rs_kc(:, :, i) Ts_kc(:, i); 0 0 0 1];
    T_wc = T_wk * T_kc;
    quat = rotm2quat(T_wc(1:3, 1:3));
    timestamp = textscan(filenames{i, 2}{1}, 'depth/%f.png');
    fprintf(fid,'%.6f %f %f %f %f %f %f %f\n', timestamp{1}, T_wc(1, 4),...
        T_wc(2, 4),T_wc(3, 4), quat(1), quat(2), quat(3), quat(4));
end
fclose(fid);








